/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.spi.processors;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Map.Entry;
import javax.annotation.processing.Filer;
import javax.annotation.processing.Messager;
import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.Types;
import javax.tools.Diagnostic.Kind;
import javax.tools.FileObject;
import javax.tools.StandardLocation;

/**
 * Utility methods shared amongst annotation processors
 * @author Evan Cowden
 */
class Util {

    /**
     * Get the path to the service file for the specified servier
     * @param serviceName
     * @return
     */
    protected static String getPath(String serviceName) {
        return String.format("META-INF/services/%s", serviceName);
    }

    /**
     * Get the qualified name for a type
     * @param typeMirror a {@code DeclaredType}
     * @return
     */
    protected static String getQualifiedName(TypeMirror typeMirror) {
        Element element = ((DeclaredType) typeMirror).asElement();
        TypeElement typeElement = (TypeElement) element;
        return typeElement.getQualifiedName().toString();
    }

    /**
     * Write SPI configuration files for all configurations in a multi map
     * @param filer
     * @param configurations the configurations, mapped service name to provider
     * name
     */
    protected static void writeConfigurations(Filer filer, SimpleMultiMap<String, String> configurations) {
        for (Entry<String, Collection<String>> entry : configurations.entrySet()) {
            writeConfiguration(filer, entry.getKey(), entry.getValue());
        }
    }

    /**
     * Write an SPI configuration file
     * @param filer
     * @param service the name of the service
     * @param providers the names of all providers for the service
     */
    protected static void writeConfiguration(Filer filer, String service, Collection<String> providers) {
        BufferedWriter out = null;
        try {
            String path = Util.getPath(service);
            FileObject file = filer.createResource(StandardLocation.CLASS_OUTPUT, "", path);
            out = new BufferedWriter(file.openWriter());

            for (String provider : providers) {
                out.write(provider);
                out.newLine();
            }

            out.flush();

        } catch (IOException ex) {
            System.out.println(String.format("Error writing configuration file for service '%s'", service));
            ex.printStackTrace();
        } finally {
            try {
                if (out != null) {
                    out.close();
                }
            } catch (IOException ex) {
                System.out.println("Error closing output writer");
                ex.printStackTrace();
            }
        }
    }

    /**
     * Verify that a provider is valid
     * @param provider the provider
     * @param service the service that the provider provides
     * @return
     */
    protected static boolean validateProvider(ProcessingEnvironment processingEnv, Element provider, TypeMirror service) {
        Messager messager = processingEnv.getMessager();

        //Verify that the provider is a class
        if (provider.getKind().equals(ElementKind.CLASS)) {
            TypeElement type = (TypeElement) provider;

            //Verify that the class is not abstract
            if (type.getModifiers().contains(Modifier.ABSTRACT)) {
                String providerName = Util.getQualifiedName(provider.asType());
                messager.printMessage(Kind.ERROR, String.format("Provider '%s' is not a valid provider. A provider must be a concrete class.", providerName), provider);
                return false;
            }

            //Verify that the provider implements the service
            Types types = processingEnv.getTypeUtils();
            if (!types.isAssignable(provider.asType(), service)) {
                String providerName = Util.getQualifiedName(provider.asType());
                String serviceName = Util.getQualifiedName(service);
                messager.printMessage(Kind.ERROR, String.format("Provider '%s' does not implement the service '%s'", providerName, serviceName), provider);
                return false;
            }


        } else {
            //The provider is not a class
            String providerName = Util.getQualifiedName(provider.asType());
            messager.printMessage(Kind.ERROR, String.format("Provider '%s' is not a valid provider. A provider must be a concrete class.", providerName), provider);
            return false;
        }

        return true;
    }
}
